from __future__ import absolute_import

from . import BaseProxy
from ..requests import munchify
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class MockChrootProxy(BaseProxy):

    def get_list(self, pagination=None):
        # TODO: implement pagination
        """List all currently available chroots.

        :return: Munch
        """
        endpoint = "/mock-chroots/list"
        response = self.request.send(endpoint=endpoint)
        return munchify(response)



# coding: utf-8
import sys


# pylint: disable=R0903
class UnicodeMixin(object):
    # http://lucumr.pocoo.org/2011/1/22/forwards-compatible-python/
    """Mixin class to handle defining the proper __str__/__unicode__
    methods in Python 2 or 3."""

    if sys.version_info[0] >= 3:  # Python 3
        def __str__(self):
            return self.__unicode__()
    else:  # Python 2
        def __str__(self):
            return self.__unicode__().encode('utf8')



from .helpers import config_from_file
from .proxies import BaseProxy
from .proxies.project import ProjectProxy
from .proxies.build import BuildProxy
from .proxies.package import PackageProxy
from .proxies.module import ModuleProxy
from .proxies.mock_chroot import MockChrootProxy
from .proxies.monitor import MonitorProxy
from .proxies.project_chroot import ProjectChrootProxy
from .proxies.build_chroot import BuildChrootProxy
from .proxies.webhook import WebhookProxy


class Client(object):
    def __init__(self, config):
        self.config = config
        self.base_proxy = BaseProxy(config)
        self.project_proxy = ProjectProxy(config)
        self.build_proxy = BuildProxy(config)
        self.package_proxy = PackageProxy(config)
        self.module_proxy = ModuleProxy(config)
        self.mock_chroot_proxy = MockChrootProxy(config)
        self.monitor_proxy = MonitorProxy(config)
        self.project_chroot_proxy = ProjectChrootProxy(config)
        self.build_chroot_proxy = BuildChrootProxy(config)
        self.webhook_proxy = WebhookProxy(config)

    @classmethod
    def create_from_config_file(cls, path=None):
        config = config_from_file(path)
        return cls(config)



from __future__ import absolute_import

import requests
from .requests import munchify

try:
    import urlparse
    from urllib import urlencode
except ImportError:
    import urllib.parse as urlparse
    from urllib.parse import urlencode


def next_page(objects):
    request = objects.__response__.request

    # Add offset to the previous request URL
    url_parts = list(urlparse.urlparse(request.url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update({"offset": objects.meta.offset + objects.meta.limit})
    url_parts[4] = urlencode(query)
    request.url = urlparse.urlunparse(url_parts)

    session = requests.Session()
    response = session.send(request)
    return munchify(response)


# @TODO remove all_pages function if unlimited generator is preferred over it
def all_pages(objects):
    all_objects = []
    while objects:
        all_objects.extend(objects)
        objects = next_page(objects)
    return all_objects


def unlimited(objects):
    if not objects:
        objects = next_page(objects)

    if not objects:
        return

    yield objects.pop()
    for item in unlimited(objects):
        yield item



from __future__ import absolute_import

from functools import wraps
import os
import time
import configparser
from munch import Munch
from .exceptions import CoprConfigException, CoprException


class List(list):
    def __init__(self, items, meta=None, response=None, proxy=None):
        list.__init__(self, items)
        self.meta = meta
        self.__response__ = response
        self.__proxy__ = proxy


def config_from_file(path=None):
    raw_config = configparser.ConfigParser()

    config = {}
    default_path = os.path.join("~", ".config", "copr")

    try:
        exists = raw_config.read(os.path.expanduser(path or default_path))
    except configparser.Error as ex:
        raise CoprConfigException(str(ex))

    if not exists:
        if path:
            # absence of the default_path is acceptable, but missing the
            # explicitly specified path= argument deserves an exception.
            raise CoprConfigException("File {0} is missing.".format(path))
        raw_config["copr-cli"] = {"copr_url": "https://copr.fedorainfracloud.org"}

    try:
        for field in ["username", "login", "token", "copr_url", "gssapi"]:
            config[field] = raw_config["copr-cli"].get(field, None)
        config["encrypted"] = raw_config["copr-cli"].getboolean("encrypted", True)
        config["gssapi"] = raw_config["copr-cli"].getboolean("gssapi")


    except configparser.Error as err:
        raise CoprConfigException("Bad configuration file: {0}".format(err))

    if config["encrypted"] and config["copr_url"].startswith("http://"):
        raise CoprConfigException("The `copr_url` should not be http, please obtain "
                                  "an up-to-date configuration from the Copr website")

    return config


def for_all_methods(decorator):
    """
    Apply a given decorator to all class methods
    """
    def decorate(cls):
        for attr in list(cls.__dict__):
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


def bind_proxy(func):
    """
    Modify a result munch and set the __proxy__ parameter
    to the actual proxy instance.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if type(result) not in [List, Munch]:
            return result
        result.__proxy__ = args[0]
        return result
    return wrapper


def wait(waitable, interval=30, callback=None, timeout=0):
    """
    Wait for a waitable thing to finish. At this point, it is possible to wait only
    for builds, but this function should be enhanced to wait for
    e.g. modules or images, etc in the future

    :param Munch/list waitable: A Munch result or list of munches
    :param int interval: How many seconds wait before requesting updated Munches from frontend
    :param callable callback: Callable taking one argument (list of build Munches).
                              It will be triggered before every sleep interval.
    :param int timeout: Limit how many seconds should be waited before this function unsuccessfully ends
    :return: list of build Munches

    Example usage:

        build1 = client.build_proxy.create_from_file(...)
        build2 = client.build_proxy.create_from_scm(...)
        wait([build1, build2])

    """
    builds = waitable if isinstance(waitable, list) else [waitable]
    watched = set([build.id for build in builds])
    munches = dict((build.id, build) for build in builds)
    failed = []
    terminate = time.time() + timeout

    while True:
        for build_id in watched.copy():
            if hasattr(munches[build_id], "__proxy__"):
                proxy = munches[build_id].__proxy__
            else:
                proxy = waitable.__proxy__
            build = munches[build_id] = proxy.get(build_id)

            if build.state in ["failed"]:
                failed.append(build_id)
            if build.state in ["succeeded", "skipped", "failed", "canceled"]:
                watched.remove(build_id)
            if build.state == "unknown":
                raise CoprException("Unknown status.")

        if callback:
            callback(list(munches.values()))
        if not watched:
            break
        if timeout and time.time() >= terminate:
            raise CoprException("Timeouted")
        time.sleep(interval)
    return list(munches.values())


def succeeded(builds):
    """
    Determine, whether the list of builds finished successfully.

    :param Munch/list builds: A list of builds or a single build Munch
    :return bool:
    """
    builds = builds if type(builds) == list else [builds]
    for build in builds:
        if build.state != "succeeded":
            return False
    return True



"""
Init file for `copr` package tests
"""

import os
import six


path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
resource_location = os.path.join(dir_path, "resources")
config_location = os.path.join(resource_location, "copr_cli.conf")

# We need to maintain python2 compatibility for EPEL7
if six.PY3:
    from unittest import mock
else:
    import mock



from __future__ import absolute_import

import os
from . import BaseProxy
from ..requests import FileRequest, munchify, POST
from ..exceptions import CoprValidationException
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class BuildProxy(BaseProxy):
    def get(self, build_id):
        """
        Return a build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/{0}".format(build_id)
        response = self.request.send(endpoint=endpoint)
        return munchify(response)

    def get_source_chroot(self, build_id):
        """
        Return a source build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/source-chroot/{0}".format(build_id)
        response = self.request.send(endpoint=endpoint)
        return munchify(response)

    def get_source_build_config(self, build_id):
        """
        Return a config for source build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/source-build-config/{0}".format(build_id)
        response = self.request.send(endpoint=endpoint)
        return munchify(response)

    def get_built_packages(self, build_id):
        """
        Return built packages (NEVRA dicts) for a given build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/built-packages/{0}".format(build_id)
        response = self.request.send(endpoint=endpoint)
        return munchify(response)

    def get_list(self, ownername, projectname, packagename=None, status=None, pagination=None):
        """
        Return a list of packages

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :param str status:
        :param pagination:
        :return: Munch
        """
        endpoint = "/build/list"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "packagename": packagename,
            "status": status,
        }
        params.update(pagination or {})

        response = self.request.send(endpoint=endpoint, params=params)
        return munchify(response)

    def cancel(self, build_id):
        """
        Cancel a build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/cancel/{0}".format(build_id)
        response = self.request.send(
            endpoint=endpoint, method=POST, auth=self.auth)
        return munchify(response)

    def create_from_urls(self, ownername, projectname, urls, buildopts=None, project_dirname=None):
        """
        Create builds from a list of URLs

        :param str ownername:
        :param str projectname:
        :param list urls:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/url"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "pkgs": urls,
            "project_dirname": project_dirname,
        }
        return self._create(endpoint, data, buildopts=buildopts)

    def create_from_url(self, ownername, projectname, url, buildopts=None, project_dirname=None):
        """
        Create a build from URL

        :param str ownername:
        :param str projectname:
        :param str url:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        if len(url.split()) > 1:
            raise CoprValidationException("This method doesn't allow submitting multiple URLs at once. "
                                          "Use `create_from_urls` instead.")
        return self.create_from_urls(ownername, projectname, [url], buildopts=buildopts,
                                     project_dirname=project_dirname)[0]

    def create_from_file(self, ownername, projectname, path, buildopts=None, project_dirname=None):
        """
        Create a build from local SRPM file

        :param str ownername:
        :param str projectname:
        :param str path:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/upload"
        f = open(path, "rb")

        data = {
            "ownername": ownername,
            "projectname": projectname,
            "project_dirname": project_dirname,
        }
        files = {
            "pkgs": (os.path.basename(f.name), f, "application/x-rpm"),
        }
        return self._create(endpoint, data, files=files, buildopts=buildopts)

    def create_from_scm(self, ownername, projectname, clone_url, committish="", subdirectory="", spec="",
                        scm_type="git", source_build_method="rpkg", buildopts=None, project_dirname=None):
        """
        Create a build from SCM repository

        :param str ownername:
        :param str projectname:
        :param str clone_url: url to a project versioned by Git or SVN
        :param str committish: name of a branch, tag, or a git hash
        :param str subdirectory: path to a subdirectory with package content
        :param str spec: path to spec file, relative to 'subdirectory'
        :param str scm_type:
        :param str source_build_method:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/scm"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "clone_url": clone_url,
            "committish": committish,
            "subdirectory": subdirectory,
            "spec": spec,
            "scm_type": scm_type,
            "source_build_method": source_build_method,
            "project_dirname": project_dirname,
        }
        return self._create(endpoint, data, buildopts=buildopts)

    def create_from_distgit(self, ownername, projectname, packagename,
                            committish=None, namespace=None, distgit=None,
                            buildopts=None, project_dirname=None):
        """
        Create a build from a DistGit repository

        :param str ownername:
        :param str projectname:
        :param str packagename: the DistGit package name to build
        :param str committish: name of a branch, tag, or a git hash
        :param str namespace: DistGit namespace, e.g. '@copr/copr' for
            the '@copr/copr/copr-cli' package
        :param str distgit: the DistGit instance name we build against,
            for example 'fedora'.  Optional, and the default is deployment
            specific.
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/distgit"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "distgit": distgit,
            "namespace": namespace,
            "package_name": packagename,
            "committish": committish,
            "project_dirname": project_dirname,
        }
        return self._create(endpoint, data, buildopts=buildopts)


    def create_from_pypi(self, ownername, projectname, pypi_package_name,
                         pypi_package_version=None, spec_template='',
                         python_versions=None, buildopts=None,
                         project_dirname=None, spec_generator=None):
        """
        Create a build from PyPI - https://pypi.org/

        :param str ownername:
        :param str projectname:
        :param str pypi_package_name:
        :param str pypi_package_version: PyPI package version (None means "latest")
        :param str spec_template: what pyp2rpm spec template to use
        :param list python_versions: list of python versions to build for
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :param str spec_generator: what tool should be used for spec generation
        :return: Munch
        """
        endpoint = "/build/create/pypi"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "pypi_package_name": pypi_package_name,
            "pypi_package_version": pypi_package_version,
            "spec_generator": spec_generator,
            "spec_template": spec_template,
            "python_versions": python_versions or [3, 2],
            "project_dirname": project_dirname,
        }
        return self._create(endpoint, data, buildopts=buildopts)

    def create_from_rubygems(self, ownername, projectname, gem_name, buildopts=None, project_dirname=None):
        """
        Create a build from RubyGems - https://rubygems.org/

        :param str ownername:
        :param str projectname:
        :param str gem_name:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/rubygems"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "gem_name": gem_name,
            "project_dirname": project_dirname,
        }
        return self._create(endpoint, data, buildopts=buildopts)

    def create_from_custom(self, ownername, projectname, script, script_chroot=None,
                           script_builddeps=None, script_resultdir=None, script_repos=None,
                           buildopts=None, project_dirname=None):
        """
        Create a build from custom script.

        :param str ownername:
        :param str projectname:
        :param script: script to execute to generate sources
        :param script_chroot: [optional] what chroot to use to generate
            sources (defaults to fedora-latest-x86_64)
        :param script_builddeps: [optional] list of script's dependencies
        :param script_resultdir: [optional] where script generates results
            (relative to cwd)
        :parm script_repos: [optional] external repositories containing script's
            dependencies
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/build/create/custom"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "script": script,
            "chroot": script_chroot,
            "builddeps": script_builddeps,
            "resultdir": script_resultdir,
            "project_dirname": project_dirname,
            "repos": script_repos,
        }
        return self._create(endpoint, data, buildopts=buildopts)

    def _create(self, endpoint, data, files=None, buildopts=None):
        data = data.copy()

        kwargs = {"api_base_url": self.api_base_url}
        if files and buildopts and "progress_callback" in buildopts:
            kwargs["progress_callback"] = buildopts["progress_callback"]
            del buildopts["progress_callback"]

        data.update(buildopts or {})
        if not files:
            response = self.request.send(
                endpoint=endpoint, data=data, method=POST, auth=self.auth)
        else:
            kwargs["files"] = files
            kwargs["connection_attempts"] = self.config.get("connection_attempts", 1)
            request = FileRequest(**kwargs)
            response = request.send(
                endpoint=endpoint, data=data, method=POST, auth=self.auth)
        return munchify(response)

    def delete(self, build_id):
        """
        Delete a build

        :param int build_id:
        :return: Munch
        """
        endpoint = "/build/delete/{0}".format(build_id)
        response = self.request.send(
            endpoint=endpoint, method=POST, auth=self.auth)
        return munchify(response)

    def delete_list(self, build_ids):
        """
        Delete multiple builds specified by a list of their ids.

        :param list[int] build_id:
        :return: Munch
        """

        endpoint = "/build/delete/list"
        data = {"builds": build_ids}
        response = self.request.send(
            endpoint=endpoint, data=data, method=POST, auth=self.auth)
        return munchify(response)



import os

from copr.v3.auth import auth_from_config
from copr.v3.requests import munchify, Request
from ..helpers import for_all_methods, bind_proxy, config_from_file


@for_all_methods(bind_proxy)
class BaseProxy(object):
    """
    Parent class for all other proxies
    """

    def __init__(self, config):
        self.config = config
        self.request = Request(
            api_base_url=self.api_base_url,
            connection_attempts=config.get("connection_attempts", 1)
        )
        self._auth = None

    @classmethod
    def create_from_config_file(cls, path=None):
        config = config_from_file(path)
        return cls(config)

    @property
    def api_base_url(self):
        return os.path.join(self.config["copr_url"], "api_3", "")

    @property
    def auth(self):
        if not self._auth:
            self._auth = auth_from_config(self.config)
        return self._auth

    def home(self):
        """
        Call the Copr APIv3 base URL

        :return: Munch
        """
        endpoint = ""
        response = self.request.send(endpoint=endpoint)
        return munchify(response)

    def auth_check(self):
        """
        Call an endpoint protected by login to check whether the user auth key is valid

        :return: Munch
        """
        endpoint = "/auth-check"
        response = self.request.send(endpoint=endpoint, auth=self.auth)
        return munchify(response)

    def auth_username(self):
        """
        Return the username (string) assigned to this configuration.  May
        contact the server and authenticate if needed.
        """
        if not self.auth.username:
            self.auth.make()
        return self.auth.username



# -*- coding: UTF-8 -*-

"""
Exceptions for Copr client.
"""


class CoprException(Exception):

    """ Basic exception class for Copr client. """
    pass


class CoprNoConfException(CoprException):

    """ Exception thrown when no config file is found. """
    pass


class CoprConfigException(CoprException):

    """ Exception thrown when the config file is incomplete or
    malformed.
    """
    pass


class CoprRequestException(CoprException):
    """ Exception thrown when the request is bad. For example,
    the user provided wrong project name or build ID.
    """
    pass


class CoprBuildException(CoprException):
    """ Exception thrown when one or more builds fail and client is waiting
    for the result.
    """
    pass


class CoprUnknownResponseException(CoprException):
    """ Exception thrown when the response is unknown to client.
    It usually means that something is broken.
    """
    pass



"""
Webhook related actions in APIv3
"""

from __future__ import absolute_import

from . import BaseProxy
from ..requests import munchify, POST
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class WebhookProxy(BaseProxy):
    """
    This class provides access to all webhook related actions in APIv3
    Methods call endpoints that starts with /api_3/webhook/
    """

    def generate(self, ownername, projectname):
        """
        Generate a new webhook secret

        :param str ownername:
        :param str projectname:
        :return: Munch
        """
        endpoint = "/webhook/generate/{ownername}/{projectname}"
        params = {
            "ownername": ownername,
            "projectname": projectname,
        }
        response = self.request.send(
            endpoint=endpoint, method=POST, params=params, auth=self.auth)
        return munchify(response)
